"""Contains various helper methods."""

import functools
import socket
import struct

DEGREE_SIGN = "\N{DEGREE SIGN}"
unpack_float = struct.Struct("<f").unpack
pack_header = struct.Struct("<BH4B").pack_into
unpack_header = struct.Struct("<BH4B").unpack_from


def crc(data: bytearray) -> int:
    """Calculates frame checksum.

    Keyword arguments:
    data - data to calculate checksum for
    """
    return functools.reduce(lambda x, y: x ^ y, data)


def to_hex(data: bytearray) -> str:
    """Converts bytearray to list of hex strings.

    Keyword arguments:
    data - data for conversion
    """
    return [f"{data[i]:02X}" for i in range(0, len(data))]


def unpack_ushort(data: bytearray) -> int:
    """Unpacks unsigned short number from bytes.

    Keyword arguments:
    data -- bytes to unpack number from
    """
    return int.from_bytes(data, byteorder="little", signed=False)


def unpack_parameter(data: bytearray, offset: int, size: int = 1) -> (int, int, int):
    """Unpacks parameter.

    Keyword arguments:
    data -- bytes to unpack number from
    offset -- data offset
    size -- parameter size in bytes
    """

    value = unpack_ushort(data[offset : offset + size])
    min_ = unpack_ushort(data[offset + size : offset + 2 * size])
    max_ = unpack_ushort(data[offset + 2 * size : offset + 3 * size])
    if check_parameter(data[offset : offset + size * 3]):
        return value, min_, max_

    return None


def ip_to_bytes(address: str) -> bytearray:
    """Converts ip address to bytes.

    Keyword arguments:
    address -- ip address as string
    """
    return socket.inet_aton(address)


def ip_from_bytes(data: bytearray) -> str:
    """Converts ip address from bytes to string representation.

    Keyword arguments:
    data -- 4 bytes to convert
    """
    return socket.inet_ntoa(data)


def merge(defaults: dict, options: dict) -> dict:
    """Merges two dictionary with options overriding defaults.

    Keyword arguments:
    defaults -- dictionary of defaults
    options -- dictionary containing options for overriding defaults
    """
    if not options:
        # Options is empty.
        return defaults

    for key in defaults.keys():
        if key not in options:
            options[key] = defaults[key]

    return options


def check_parameter(data: bytearray) -> bool:
    """Checks if parameter contains any bytes besides 0xFF.

    Keyword arguments:
    data -- parameter bytes
    """
    for byte in data:
        if byte != 0xFF:
            return True

    return False


def uid_stamp(message: str) -> str:
    """Calculates UID stamp.

    Keyword arguments:
    message -- uid message
    """
    crc_ = 0xA3A3
    for byte in message:
        int_ = ord(byte)
        crc_ = crc_ ^ int_
        for _ in range(8):
            if crc_ & 1:
                crc_ = (crc_ >> 1) ^ 0xA001
                continue

            crc_ = crc_ >> 1

    return chr(crc_ % 256) + chr((crc_ // 256) % 256)


def uid_5bits_to_char(number: int) -> str:
    """Converts 5 bits from UID to ASCII character.

    Keyword arguments:
    number -- byte for conversion
    """
    if number < 0 or number >= 32:
        return "#"

    if number < 10:
        return chr(ord("0") + number)

    char = chr(ord("A") + number - 10)

    return "Z" if char == "O" else char


def make_list(data: dict, include_keys: bool = True):
    """Converts dictionary to string.

    Keyword arguments:
    data -- dictionary to convert to string
    include_keys -- determines if keys should be included
    """

    output = ""
    for k, v in data.items():
        output += f"- {k}: {v}\n" if include_keys else f"- {v}\n"

    return output.rstrip()



"""Contains econet connection shortcut."""

from .econet import SerialConnection, TcpConnection
from .version import __version__  # noqa


def econet_tcp_connection(host: str, port: int, **kwargs):
    """Creates TCP connection instance.

    Keyword arguments:
    host -- serial device ip/hostname
    port -- serial device port
    **kwargs -- keyword arguments directly passed to asyncio's
        create_connection method
    """
    return TcpConnection(host, port, **kwargs)


def econet_serial_connection(device: str, baudrate: int = 115200, **kwargs):
    """Creates serial connection instance.

    Keyword arguments:
    device -- serial device url, e. g. /dev/ttyUSB0
    **kwargs -- keyword arguments directly passed to asyncio's
        create_connection method
    """
    return SerialConnection(device=device, baudrate=baudrate, **kwargs)



"""Contains reader and writer classes."""

from __future__ import annotations

from asyncio import StreamReader, StreamWriter

from . import util
from .constants import (
    BROADCAST_ADDRESS,
    ECONET_ADDRESS,
    HEADER_SIZE,
    READER_BUFFER_SIZE,
)
from .exceptions import ChecksumError, LengthError
from .factory import FrameFactory
from .frame import Frame, Request


class FrameWriter:
    """Used to asynchronously write frames to a connection using
    asyncio's StreamWriter and maintains write queue.
    """

    def __init__(self, writer: StreamWriter):
        """Creates instance of FrameWriter.

        Keyword arguments:
        writer -- instance of asyncio.StreamWriter
        """
        self.writer = writer
        self._queue = []

    def __len__(self) -> int:
        """Gets write queue length."""
        return len(self._queue)

    def queue(self, *frames: list[Frame]) -> None:
        """Adds frame to write queue.

        Keyword arguments:
        frame -- Frame instance to add
        """
        for frame in frames:
            if isinstance(frame, Frame):
                self._queue.append(frame)

    def queue_is_empty(self) -> bool:
        """Checks if write queue is empty."""
        return len(self._queue) == 0

    def collect(self, requests: dict[Request]) -> None:
        """Collects changed parameters and adds them to write queue.

        Keyword arguments:
        parameters -- list of device parameters
        """
        for request in requests:
            self.queue(request)

    async def process_queue(self) -> None:
        """Processes top-most write request from the stack."""
        if self._queue:
            frame = self._queue.pop(0)
            await self.write(frame)

    async def write(self, frame: Frame) -> None:
        """Writes frame to connection and waits for buffer to drain.

        Keyword arguments:
        frame -- Frame instance to add
        """
        self.writer.write(frame.to_bytes())
        await self.writer.drain()

    async def close(self) -> None:
        """Closes StreamWriter."""
        self.writer.close()
        await self.writer.wait_closed()


class FrameReader:
    """Used to read and parse received frames
    using asyncio's StreamReader.
    """

    def __init__(self, reader: StreamReader):
        """Creates FrameReader instance.

        Keyword arguments:
        reader -- instance of asyncio.StreamReader
        """
        self.reader = reader

    async def read(self) -> Frame:
        """Attempts to read READER_BUFFER_SIZE bytes, find
        valid frame in it and return corresponding Frame instance.
        """
        buffer = await self.reader.read(READER_BUFFER_SIZE)

        if len(buffer) >= HEADER_SIZE:
            header = buffer[0:HEADER_SIZE]
            [
                _,
                length,
                recipient,
                sender,
                sender_type,
                econet_version,
            ] = util.unpack_header(header)

            if recipient in [ECONET_ADDRESS, BROADCAST_ADDRESS]:
                # Destination address is econet or broadcast.
                payload = buffer[HEADER_SIZE:length]

                if HEADER_SIZE + len(payload) != length:
                    raise LengthError()

                if payload[-2] != util.crc(header + payload[:-2]):
                    raise ChecksumError()

                return FrameFactory().get_frame(
                    type_=payload[0],
                    recipient=recipient,
                    message=payload[1:-2],
                    sender=sender,
                    sender_type=sender_type,
                    econet_version=econet_version,
                )



"""Contains mixer parameter structure parser."""

from pyplumio import util
from pyplumio.constants import DATA_MIXERS, MIXER_PARAMS


def from_bytes(message: bytearray, offset: int = 0, data: dict = None) -> (dict, int):
    """Parses frame message into usable data.

    Keyword arguments:
    message -- ecoNET message
    offset -- current data offset
    """
    if data is None:
        data = {}

    first_parameter = message[1]
    parameters_number = message[2]
    mixers_number = message[3]
    total_parameters = mixers_number * parameters_number
    offset = 4
    data[DATA_MIXERS] = []
    if parameters_number == 0:
        return data, offset

    mixer = {}
    for index in range(first_parameter, total_parameters + first_parameter):
        parameter = util.unpack_parameter(message, offset)
        if parameter is not None:
            mixer[f"{MIXER_PARAMS[index%mixers_number]}"] = parameter

        if index % parameters_number == 0:
            data[DATA_MIXERS].append(mixer)
            mixer = {}

        offset += 3

    return data, offset



"""Contains singleton pattern implementation."""
from __future__ import annotations


class Singleton:
    """Singleton pattern implementation."""

    def __new__(cls: str) -> Singleton:
        """Gets singleton instance.

        Keyword arguments:
        cls - current class name
        """
        if not hasattr(cls, "instance"):
            cls.instance = super(Singleton, cls).__new__(cls)

        return cls.instance



"""Contains constants."""

FRAME_START: int = 0x68
FRAME_END: int = 0x16
HEADER_SIZE: int = 7
BROADCAST_ADDRESS: int = 0x00
ECONET_ADDRESS: int = 0x56
ECONET_TYPE: int = 0x30
ECONET_VERSION: int = 0x05

READER_BUFFER_SIZE: int = 1000
READER_TIMEOUT: int = 5
RECONNECT_TIMEOUT: int = 30

UID_BASE: int = 32
UID_BASE_BITS: int = 5
UID_CHAR_BITS: int = 8

WLAN_ENCRYPTION_UNKNOWN = 0
WLAN_ENCRYPTION_NONE = 1
WLAN_ENCRYPTION_WEP = 2
WLAN_ENCRYPTION_WPA = 3
WLAN_ENCRYPTION_WPA2 = 4
WLAN_ENCRYPTION = (
    WLAN_ENCRYPTION_UNKNOWN,
    WLAN_ENCRYPTION_NONE,
    WLAN_ENCRYPTION_WEP,
    WLAN_ENCRYPTION_WPA,
    WLAN_ENCRYPTION_WPA2,
)

DEFAULT_IP = "0.0.0.0"
DEFAULT_NETMASK = "255.255.255.0"

MODULE_PANEL: str = "module_panel"
MODULE_A: str = "module_a"
MODULE_B: str = "module_b"
MODULE_C: str = "module_c"
MODULE_LAMBDA: str = "module_lambda"
MODULE_ECOSTER: str = "module_ecoster"
MODULES: list = (
    MODULE_PANEL,
    MODULE_A,
    MODULE_B,
    MODULE_C,
    MODULE_LAMBDA,
    MODULE_ECOSTER,
)

MODES: list = (
    "Off",
    "Starting",
    "Kindling",
    "Heating",
    "Supervision",
    "Cooling",
    "Standby",
)

MIXER_TEMP: str = "temp"
MIXER_TARGET: str = "target"
MIXER_PUMP: str = "pump"
MIXER_DATA: list = (
    MIXER_TEMP,
    MIXER_TARGET,
    MIXER_PUMP,
)

ECOSTER_CONTACTS: str = "contacts"
ECOSTER_SCHEDULE: str = "schedule"
ECOSTER_MODE: str = "mode"
ECOSTER_TEMP: str = "temp"
ECOSTER_TARGET: str = "target"

DATA_ALARMS: str = "alarms"
DATA_CO_STATUS: str = "co_status"
DATA_CO_TARGET: str = "co_target"
DATA_CWU_STATUS: str = "cwu_status"
DATA_CWU_TARGET: str = "cwu_target"
DATA_FAN_POWER: str = "fan_power"
DATA_FRAMES: str = "frames"
DATA_FUEL_CONSUMPTION: str = "fuel_consumption"
DATA_FUEL_LEVEL: str = "fuel_level"
DATA_LAMBDA_LEVEL: str = "lambda_level"
DATA_LAMBDA_STATUS: str = "lambda_status"
DATA_LAMBDA_TARGET: str = "lambda_target"
DATA_MIXERS: str = "mixers"
DATA_MODE: str = "mode"
DATA_POWER: str = "power"
DATA_LOAD: str = "load"
DATA_THERMOSTAT: str = "thermostat"
DATA_THERMOSTATS: str = "thermostats"
DATA_TRANSMISSION: str = "transmission"

TEMPERATURES: list = (
    "co_temp",
    "feeder_temp",
    "cwu_temp",
    "outside_temp",
    "back_temp",
    "exhaust_temp",
    "optical_temp",
    "upper_buffer_temp",
    "lower_buffer_temp",
    "upper_solar_temp",
    "lower_solar_temp",
    "fireplace_temp",
    "total_gain",
    "hydraulic_coupler_temp",
    "exchanger_temp",
    "air_in_temp",
    "air_out_temp",
)

OUTPUTS: list = (
    "fan",
    "feeder",
    "co_pump",
    "cwu_pump",
    "ciculation_pump",
    "lighter",
    "alarm",
    "outer_boiler",
    "fan2_exhaust",
    "feeder2",
    "outer_feeder",
    "solar_pump",
    "fireplace_pump",
    "gcz_contact",
    "blow_fan1",
    "blow_fan2",
)

DATA_CO_PUMP_FLAG: str = "co_pump_flag"
DATA_CWU_PUMP_FLAG: str = "cwu_pump_flag"
DATA_CIRCULATION_PUMP_FLAG: str = "circulation_pump_flag"
DATA_SOLAR_PUMP_FLAG: str = "solar_pump_flag"
FLAGS: list = (
    DATA_CO_PUMP_FLAG,
    DATA_CWU_PUMP_FLAG,
    DATA_CIRCULATION_PUMP_FLAG,
    DATA_SOLAR_PUMP_FLAG,
)

STATUSES: list = (
    "co_target",
    "co_status",
    "cwu_target",
    "cwu_status",
)

DEVICE_DATA: list = [
    DATA_ALARMS,
    DATA_FAN_POWER,
    DATA_FRAMES,
    DATA_FUEL_CONSUMPTION,
    DATA_FUEL_LEVEL,
    DATA_LAMBDA_LEVEL,
    DATA_LAMBDA_STATUS,
    DATA_LAMBDA_TARGET,
    DATA_LOAD,
    DATA_MODE,
    DATA_POWER,
    DATA_THERMOSTAT,
    DATA_THERMOSTATS,
    DATA_TRANSMISSION,
]
DEVICE_DATA.extend(TEMPERATURES)
DEVICE_DATA.extend(OUTPUTS)
DEVICE_DATA.extend(FLAGS)
DEVICE_DATA.extend(STATUSES)
DEVICE_DATA.extend(MODULES)

DEVICE_PARAMS: list = (
    "airflow_power_100",
    "airflow_power_50",
    "airflow_power_30",
    "power_100",
    "power_50",
    "power_30",
    "max_fan_boiler_power",
    "min_fan_boiler_power",
    "fuel_feeding_time_100",
    "fuel_feeding_time_50",
    "fuel_feeding_time_30",
    "fuel_feeding_break_100",
    "fuel_feeding_break_50",
    "fuel_feeding_break_30",
    "cycle_time",
    "h2_hysteresis",
    "h1_hysteresis",
    "boiler_hysteresis",
    "control_mode",
    "min_fl_power",
    "max_fl_power",
    "min_boiler_power",
    "max_boiler_power",
    "min_fan_power",
    "max_fan_power",
    "t_reduction_airflow",
    "fan_power_gain",
    "fuel_flow_correction_fl",
    "fuel_flow_correction",
    "airflow_correction_100",
    "feeder_correction_100",
    "airflow_correction_50",
    "feeder_correction_50",
    "airflow_correction_30",
    "feeder_correction_30",
    "airflow_power_grate",
    "hist_boiler_grate",
    "supervision_work_airflow",
    "supervision_work_airflow_brake",
    "co_temp_grate",
    "det_time_fuel_grate",
    "airflow_power_kindle",
    "small_airflow_power_kindle",
    "airflow_kindle_delay",
    "scavenge_kindle",
    "feeder_kindle",
    "feeder_kindle_weight",
    "kindle_time",
    "warming_up_time",
    "fumes_temp_kindle_finish",
    "finish_kindle_threshold",
    "fumes_delta_kindle",
    "delta_t_kindle",
    "min_kindle_power_time",
    "scavenge_after_kindle",
    "airflow_power_after_kindle",
    "supervision_time",
    "feed_time_supervision",
    "feed_time_supervision_weight",
    "feed_supervision_break",
    "supervision_cycle_duration",
    "airflow_power_supervision",
    "fan_supervison_break",
    "fan_work_supervision",
    "increase_fan_support_mode",
    "max_extinguish_time",
    "min_extinguish_time",
    "extinguish_time",
    "airflow_power_extinguish",
    "airflow_work_extinguish",
    "airflow_brake_extinguish",
    "scavenge_start_extinguish",
    "scavenge_stop_extinguish",
    "clean_begin_time",
    "extinguish_clean_time",
    "airflow_power_clean",
    "warming_up_brake_time",
    "warming_up_cycle_time",
    "remind_time",
    "lambda_work",
    "lambda_correction_range",
    "oxygen_100",
    "oxygen_50",
    "oxygen_30",
    "oxygen_correction_fl",
    "fuel_kg_h",
    "feeder_calibration",
    "fuel_factor",
    "calorific_kwh_kg",
    "fuel_detection_time",
    "fumes_temp_fuel_detection",
    "schedule_feeder_2",
    "feed2_h1",
    "feed2_h2",
    "feed2_h3",
    "feed2_h4",
    "feed2_work",
    "feed2_break",
    "co_set_temp",
    "min_co_set_temp",
    "max_co_set_temp",
    "switch_co_temp",
    "pause_co_cwu",
    "pause_term",
    "work_term",
    "increase_temp_co",
    "program_control_co",
    "co_heat_curve",
    "parallel_co_heat_curve",
    "weather_factor",
    "term_boiler_operation",
    "term_boiler_mode",
    "decrease_set_co_term",
    "term_pump_off",
    "al_boiler_temp",
    "max_feed_temp",
    "extern_boiler_temp",
    "alarm_notif",
    "pump_hysteresis",
    "cwu_set_temp",
    "min_cwu_set_temp",
    "max_cwu_set_temp",
    "cwu_work_mode",
    "cwu_hysteresis",
    "cwu_disinfection",
    "auto_summer",
    "summer_temp_on",
    "summer_temp_off",
    "cwu_feeding_extension",
    "circulation_control",
    "circulation_pause_time",
    "circulation_work_time",
    "circulation_start_temp",
    "buffer_control",
    "buffer_max_temp",
    "min_buffer_temp",
    "buffer_histeresis",
    "buffer_load_start",
    "buffer_load_stop",
    "boiler_control",
)

MIXER_PARAMS: list = (
    "mix_set_temp",
    "min_mix_set_temp",
    "max_mix_set_temp",
    "low_mix_set_temp",
    "ctrl_weather_mix",
    "mix_heat_curve",
    "parallel_offset_heat_curv",
    "weather_temp_factor",
    "mix_operation",
    "mix_insensitivity",
    "mix_therm_operation",
    "mix_therm_mode",
    "mix_off_therm_pump",
    "mix_summer_work",
)



"""Contains temperatures structure parser."""

import math

from pyplumio import util
from pyplumio.constants import TEMPERATURES


def from_bytes(message: bytearray, offset: int = 0, data: dict = None) -> (dict, int):
    """Parses frame message into usable data.

    Keyword arguments:
    message -- ecoNET message
    offset -- current data offset
    """
    if data is None:
        data = {}

    temp_number = message[offset]
    offset += 1
    for _ in range(temp_number):
        index = message[offset]
        temp = util.unpack_float(message[offset + 1 : offset + 5])[0]

        if (not math.isnan(temp)) and 0 <= index < len(TEMPERATURES):
            # Temperature exists and index is in the correct range.
            data[TEMPERATURES[index]] = temp

        offset += 5

    return data, offset



"""Contains lambda structure parser."""

import math

from pyplumio import util
from pyplumio.constants import DATA_LAMBDA_LEVEL, DATA_LAMBDA_STATUS, DATA_LAMBDA_TARGET


def from_bytes(message: bytearray, offset: int = 0, data: dict = None) -> (dict, int):
    """Parses frame message into usable data.

    Keyword arguments:
    message -- ecoNET message
    offset -- current data offset
    """
    if data is None:
        data = {}

    if message[offset] == 0xFF:
        offset += 1
        return data, offset

    data[DATA_LAMBDA_STATUS] = message[offset]
    data[DATA_LAMBDA_TARGET] = message[offset + 1]
    lambda_level = util.unpack_ushort(message[offset + 2 : offset + 4])
    if math.isnan(lambda_level):
        lambda_level = None

    data[DATA_LAMBDA_LEVEL] = lambda_level
    offset += 4

    return data, offset



"""Contains output flags structure parser."""

from pyplumio import util
from pyplumio.constants import (
    DATA_CIRCULATION_PUMP_FLAG,
    DATA_CO_PUMP_FLAG,
    DATA_CWU_PUMP_FLAG,
    DATA_SOLAR_PUMP_FLAG,
)


def from_bytes(message: bytearray, offset: int = 0, data: dict = None) -> (dict, int):
    """Parses frame message into usable data.

    Keyword arguments:
    message -- ecoNET message
    offset -- current data offset
    """
    if data is None:
        data = {}

    output_flags = util.unpack_ushort(message[offset : offset + 4])
    data[DATA_CO_PUMP_FLAG] = bool(output_flags & 0x004)
    data[DATA_CWU_PUMP_FLAG] = bool(output_flags & 0x008)
    data[DATA_CIRCULATION_PUMP_FLAG] = bool(output_flags & 0x010)
    data[DATA_SOLAR_PUMP_FLAG] = bool(output_flags & 0x800)
    offset += 4

    return data, offset



"""Contains classes for mixer support."""
from __future__ import annotations

from .constants import MIXER_DATA, MIXER_PARAMS
from .helpers.base_device import BaseDevice
from .helpers.parameter import Parameter


class Mixer(BaseDevice):
    """Mixer device representation."""

    def __init__(self, data: dict = None, parameters: dict = None, index: int = 0):
        """Creates device instance.

        Keyword arguments:
        data -- device data
        parameters -- editable parameters
        """
        super().__init__(data, parameters)
        self._index = index

    def set_data(self, data: dict) -> None:
        """Sets mixer data.

        Keyword arguments:
        data -- mixer immutable attributes
        """
        for name, value in data.items():
            if name in MIXER_DATA:
                self._data[name] = value

    def set_parameters(self, parameters: dict) -> None:
        """Sets mixer parameters.

        Keyword arguments:
        parameters -- mixer changeable parameters
        """
        for name, parameter in parameters.items():
            if name in MIXER_PARAMS:
                self._parameters[name] = Parameter(name, *parameter, self._index)

    @property
    def editable_parameters(self) -> list:
        """Returns list of editable parameters."""
        return MIXER_PARAMS


class MixersCollection:
    """Collection of mixer devices."""

    def __init__(self, mixers: list = None):
        """Creates mixer collection instance.

        Keyword arguments:
        data -- device data
        parameters -- editable parameters
        """
        if mixers is None:
            self._mixers = []
        else:
            self._mixers = mixers

    def __repr__(self) -> str:
        """Returns serializable string representation of the class."""
        return f"""MixersCollection(
    mixers = {self._mixers}
)
"""

    def __str__(self) -> str:
        """Returns string representation of the class."""
        output = ""
        for index, mixer in enumerate(self._mixers):
            mixer_string = mixer.__str__().replace("\n", "\n    ")
            output += f"- {index}:\n    {mixer_string}\n\n"

        return output

    def __len__(self) -> int:
        """Returns number of mixers in collection."""
        return len(self._mixers)

    def __call__(self, index: int) -> Mixer | None:
        """Returns mixer instance by index.

        Keyword arguments:
        index -- mixer index
        """
        try:
            return self._mixers[index]
        except IndexError:
            return None

    def set_data(self, mixers: list):
        """Creates mixer instance if not exists and sets mixers data.

        Keyword arguments:
        mixers -- list of mixer data
        """
        for index, data in enumerate(mixers):
            try:
                self._mixers[index].set_data(data)
            except IndexError:
                self._mixers.append(Mixer(data=data, index=index))

    def set_parameters(self, mixers: list):
        """Creates mixer instance if not exists and
        sets mixers parameters.

        Keyword arguments:
        mixers -- list of mixer parameters
        """
        for index, parameters in enumerate(mixers):
            try:
                self._mixers[index].set_parameters(parameters)
            except IndexError:
                self._mixers.append(Mixer(parameters=parameters, index=index))

    @property
    def mixers(self):
        """Returns list of mixers."""
        return self._mixers

    @property
    def queue(self):
        """Clears and returns changed parameters queue
        for every mixer.
        """
        queue = []
        for mixer in self._mixers:
            queue.extend(mixer.queue)

        return queue



"""Contains exceptions."""


class ChecksumError(Exception):
    """Raised on checksum error while parsing frame content."""


class LengthError(Exception):
    """Raised on unexpected frame length while parsing frame content."""


class FrameTypeError(Exception):
    """Raised on unknown frame type."""


class UninitializedParameterError(Exception):
    """Raised on uninitialized device parameter."""



"""Contains base device class."""
from __future__ import annotations

from abc import ABC, abstractmethod

from pyplumio import util
from pyplumio.exceptions import UninitializedParameterError
from pyplumio.frame import Request

from .parameter import Parameter


class BaseDevice(ABC):
    """Base device class."""

    def __init__(self, data: dict = None, parameters: dict = None):
        """Creates device instance.

        Keyword arguments:
        data -- device data
        parameters -- editable parameters
        """
        self.__dict__["_data"] = {}
        self.__dict__["_parameters"] = {}

        if data is not None:
            self.set_data(data)

        if parameters is not None:
            self.set_parameters(parameters)

        self._queue = []

    def __getattr__(self, name: str):
        """Gets data or parameter as class attribute.

        Keyword arguments:
        name -- name of property to get
        """
        key = name.lower()
        if key in self._data:
            return self._data[key]

        if key in self._parameters:
            return self._parameters[key]

        return None

    def __setattr__(self, name: str, value) -> None:
        """Sets class attribute or device parameter.

        Keyword arguments:
        name -- attribute name
        value -- attribute value
        """
        key = name.lower()
        if key in self._data:
            raise AttributeError()

        if key not in self.editable_parameters:
            self.__dict__[name] = value

        elif key in self._parameters:
            self._parameters[key].set(value)
            self._queue.append(self._parameters[key].request)

        else:
            raise UninitializedParameterError()

    def __repr__(self) -> str:
        """Returns serializable string representation of the class."""
        return f"""{self.__class__.__name__}(
    data = {self._data},
    parameters = {self._parameters}
)
""".strip()

    def __str__(self) -> str:
        """Returns string representation of the class."""
        return f"""
Data:
{util.make_list(self._data)}

Parameters:
{util.make_list(self._parameters, include_keys = False)}
""".strip()

    @property
    def queue(self) -> list[Request]:
        """Clears and returns changed parameters queue."""
        queue = self._queue
        self._queue = []
        return queue

    @property
    def data(self) -> dict:
        """Returns device data."""
        return self._data

    @property
    def parameters(self) -> dict[Parameter]:
        """Returns device parameters."""
        return self._parameters

    @abstractmethod
    def set_data(self, data: dict) -> None:
        """Sets device data.

        Keyword arguments:
        data -- device immutable attributes
        """

    @abstractmethod
    def set_parameters(self, parameters: dict) -> None:
        """Sets device parameters.

        Keyword arguments:
        parameters -- device changeable parameters
        """

    @property
    @abstractmethod
    def editable_parameters(self) -> list:
        """Returns list of editable parameters."""



"""Contains alarms structure parser."""

from pyplumio.constants import DATA_ALARMS


def from_bytes(message: bytearray, offset: int = 0, data: dict = None) -> (dict, int):
    """Parses frame message into usable data.

    Keyword arguments:
    message -- ecoNET message
    offset -- current data offset
    """
    if data is None:
        data = {}

    alarms = []
    alarms_number = message[offset]
    for i in range(alarms_number):
        alarms.append(message[offset + i])

    offset += alarms_number + 1
    data[DATA_ALARMS] = alarms

    return data, offset



"""Contains variable string structure parser."""


def from_bytes(message: bytearray, offset: int = 0) -> (str, int):
    """Parses frame message into usable data.

    Keyword arguments:
    message -- ecoNET message
    offset -- current data offset
    """
    string_length = message[offset]
    offset += 1
    string = message[offset : offset + string_length + 1].decode()
    offset += string_length + 1

    return string, offset



"""Contains thermostats structure parser."""

from pyplumio import util
from pyplumio.constants import (
    DATA_THERMOSTATS,
    ECOSTER_CONTACTS,
    ECOSTER_MODE,
    ECOSTER_SCHEDULE,
    ECOSTER_TARGET,
    ECOSTER_TEMP,
)


def from_bytes(message: bytearray, offset: int = 0, data: dict = None) -> (list, int):
    """Parses frame message into usable data.

    Keyword arguments:
    message -- ecoNET message
    offset -- current data offset
    """
    if data is None:
        data = {}

    if message[offset] == 0xFF:
        offset += 1
        return data, offset

    thermostats = []
    therm_contacts = message[offset]
    offset += 1
    therm_number = message[offset]
    offset += 1
    if therm_number > 0:
        contact_mask = 1
        schedule_mask = 1 << 3
        for therm in range(1, therm_number + 1):
            therm = {}
            therm[ECOSTER_CONTACTS] = bool(therm_contacts & contact_mask)
            therm[ECOSTER_SCHEDULE] = bool(therm_contacts & schedule_mask)
            therm[ECOSTER_MODE] = bool(message[offset])
            therm[ECOSTER_TEMP] = util.unpack_float(message[offset + 1 : offset + 5])[0]
            therm[ECOSTER_TARGET] = util.unpack_float(message[offset + 5 : offset + 9])[
                0
            ]
            thermostats.append(therm)
            offset += 9
            contact_mask = contact_mask << 1
            schedule_mask = schedule_mask << 1

    data[DATA_THERMOSTATS] = thermostats

    return data, offset



"""Contains frame versions structure parser."""

from pyplumio import util
from pyplumio.constants import DATA_FRAMES


def from_bytes(message: bytearray, offset: int = 0, data: dict = None) -> (list, int):
    """Parses frame message into usable data.

    Keyword arguments:
    message -- ecoNET message
    offset -- current data offset
    """
    if data is None:
        data = {}

    versions = {}

    frames_number = message[offset]
    offset += 1
    for _ in range(frames_number):
        frame_type = message[offset]
        version = util.unpack_ushort(message[offset + 1 : offset + 3])
        versions[frame_type] = version
        offset += 3

    data[DATA_FRAMES] = versions

    return data, offset



"""Contains class that are used for data storage."""

from __future__ import annotations

from .exceptions import FrameTypeError
from .factory import FrameFactory
from .frame import Request


class FrameBucket:
    """Keeps track of frame versions and stores versioning data."""

    def __init__(self, versions: dict = None):
        """Created FrameBucket instance.

        Keyword arguments:
        versions -- dictionary containing frame versions
        """
        self.versions = {}
        self._queue = []

        if versions is not None:
            self.versions = versions

    def __len__(self) -> int:
        """Gets number of versioned frames."""
        return len(self.versions)

    def __repr__(self) -> str:
        """Returns serializable string representation."""
        return f"""FrameBucket(
    versions: {self.versions}
)
"""

    def fill(self, frames: dict) -> None:
        """Fills storage with frame versions.

        Keyword arguments:
        frames -- dictionary of frames keyed by frame versions
        """
        for type_, version in frames.items():
            if type_ not in self.versions or self.versions[type_] != version:
                # We don't have this frame or it's version has changed.
                self.update(type_, version)

    def update(self, type_: int, version: int) -> None:
        """Schedules frame update.

        Keyword arguments:
        type_ -- type of frame
        version -- new frame version to update to
        """
        try:
            frame = FrameFactory().get_frame(type_=type_)
        except FrameTypeError:
            return None

        if not isinstance(frame, Request):
            # Do not process responses.
            return None

        self.versions[type_] = version
        self._queue.append(frame)
        return None

    @property
    def queue(self) -> list[Request]:
        """Clears and returns changed frames queue."""
        queue = self._queue
        self._queue = []
        return queue



"""Contains device parameter representation."""

from pyplumio.constants import MIXER_PARAMS
from pyplumio.frame import Request
from pyplumio.frames.requests import BoilerControl, SetMixerParameter, SetParameter


class Parameter:
    """Device parameter representation."""

    def __init__(self, name: str, value: int, min_: int, max_: int, extra=None):
        """Creates parameter."""
        self.name = name
        self.value = int(value)
        self.min_ = min_
        self.max_ = max_
        self.extra = extra

    def set(self, value) -> None:
        """Sets parameter value.

        Keyword arguments:
        value -- new value to set parameter to
        """
        if self.value != value and self.min_ <= value <= self.max_:
            self.value = value

    @property
    def request(self) -> Request:
        """Returns request to change parameter."""
        if self.name == "BOILER_CONTROL":
            return BoilerControl(data=self.__dict__)

        if self.name in MIXER_PARAMS:
            return SetMixerParameter(data=self.__dict__)

        return SetParameter(data=self.__dict__)

    def __repr__(self) -> str:
        """Returns serializable string representation."""
        return f"""Parameter(
    name = {self.name},
    value = {self.value},
    min_ = {self.min_},
    max_ = {self.max_},
    extra = {self.extra}
)""".strip()

    def __str__(self) -> str:
        """Returns string representation."""
        return f"{self.name}: {self.value} (range {self.min_} - {self.max_})"

    def __int__(self) -> int:
        """Returns integer representation of parameter value.

        Keyword arguments:
        other -- other value to compare to
        """
        return int(self.value)

    def __eq__(self, other) -> int:
        """Compares if parameter value is equal to other.

        Keyword arguments:
        other -- other value to compare to
        """
        return self.value == other

    def __ge__(self, other) -> int:
        """Compares if parameter value is greater or equal to other.

        Keyword arguments:
        other -- other value to compare to
        """
        return self.value >= other

    def __gt__(self, other) -> int:
        """Compares if parameter value is greater than other.

        Keyword arguments:
        other -- other value to compare to
        """
        return self.value > other

    def __le__(self, other) -> int:
        """Compares if parameter value is less or equal to other.

        Keyword arguments:
        other -- other value to compare to
        """
        return self.value <= other

    def __lt__(self, other) -> int:
        """Compares if parameter value is less that other.

        Keyword arguments:
        other -- other value to compare to
        """
        return self.value < other



"""Contains outputs structure parser."""

import math

from pyplumio import util
from pyplumio.constants import OUTPUTS


def from_bytes(message: bytearray, offset: int = 0, data: dict = None) -> (dict, int):
    """Parses frame message into usable data.

    Keyword arguments:
    message -- ecoNET message
    offset -- current data offset
    """
    if data is None:
        data = {}

    outputs = util.unpack_ushort(message[offset : offset + 4])
    for index, output in enumerate(OUTPUTS):
        data[output] = bool(outputs & int(math.pow(2, index)))

    offset += 4

    return data, offset



"""Contains request frame classes."""

from pyplumio.constants import DEVICE_PARAMS, MIXER_PARAMS
from pyplumio.frame import Request

from . import responses


class ProgramVersion(Request):
    """ProgramVersion requests version info from ecoMAX device."""

    type_: int = 0x40

    def response(self, **args):
        """Returns corresponding response frame instance.

        Keyword arguments:
        args -- arguments to pass to response frame constructor
        """
        return responses.ProgramVersion(recipient=self.sender, **args)


class CheckDevice(Request):
    """CheckDevice requests if device is available."""

    type_: int = 0x30

    def response(self, **args):
        """Returns corresponding response frame instance.

        Keyword arguments:
        args -- arguments to pass to response frame constructor
        """
        return responses.DeviceAvailable(recipient=self.sender, **args)


class UID(Request):
    """Requests device UID."""

    type_: int = 0x39


class Password(Request):
    """Requests service password."""

    type_: int = 0x3A


class Timezones(Request):
    """Requests timezones."""

    type_: int = 0x36


class Parameters(Request):
    """Requests current editable parameters."""

    type_: int = 0x31

    def create_message(self) -> bytearray:
        """Creates SetParameter message."""

        message = bytearray()
        message.append(0xFF)  # Number of parameters.
        message.append(0x00)  # Index of parameters.
        return message


class MixerParameters(Request):
    """Requests current mixer parameters."""

    type_: int = 0x32


class DataStructure(Request):
    """Requests current regulator data structure."""

    type_: int = 0x55


class SetParameter(Request):
    """Changes current regulator parameter."""

    type_: int = 0x33

    def create_message(self) -> bytearray:
        """Creates SetParameter message."""

        message = bytearray()
        name = self._data["name"]
        value = self._data["value"]
        if name in DEVICE_PARAMS:
            message.append(DEVICE_PARAMS.index(name))
            message.append(value)

        return message


class SetMixerParameter(Request):
    """Sets mixer parameter."""

    type_: int = 0x34

    def create_message(self) -> bytearray:
        """Creates SetParameter message."""

        message = bytearray()
        name = self._data["name"]
        value = self._data["value"]
        index = self._data["extra"]
        if name in MIXER_PARAMS:
            message.append(index)
            message.append(MIXER_PARAMS.index(name))
            message.append(value)

        return message


class BoilerControl(Request):
    """Changes regulator state (on/off)."""

    type_: int = 0x3B

    def create_message(self) -> bytearray:
        """Creates BoilerControl message."""

        message = bytearray()
        message.append(self._data["value"])

        return message



"""Contains frame class."""

from __future__ import annotations

from abc import ABC, abstractmethod

from . import util
from .constants import (
    BROADCAST_ADDRESS,
    ECONET_ADDRESS,
    ECONET_TYPE,
    ECONET_VERSION,
    FRAME_END,
    FRAME_START,
    HEADER_SIZE,
)


class Frame(ABC):
    """Used as base class for creating and parsing request and response
    frames.
    """

    def __init__(
        self,
        type_: int = None,
        recipient: int = BROADCAST_ADDRESS,
        message: bytearray = bytearray(),
        sender: int = ECONET_ADDRESS,
        sender_type: int = ECONET_TYPE,
        econet_version: int = ECONET_VERSION,
        data=None,
    ):
        """Creates new Frame object.

        Keyword arguments:
            type_ -- integer repsentation of frame type
            recipient -- integer repsentation of recipient address
            message -- frame body as bytearray
            sender -- integer respresentation of sender address
            sender_type -- sender type
            econet_version -- version of econet protocol
            data -- frame data, that is used to construct frame message
        """
        self._data = data
        self.recipient = recipient
        self.sender = sender
        self.sender_type = sender_type
        self.econet_version = econet_version
        if type_ is not None:
            self.type_ = type_

        if not message:
            # If frame message is not passed in constructor.
            try:
                self.message = self.create_message()
            except NotImplementedError:
                self.message = bytearray()

        else:
            self.message = message

    def __repr__(self) -> str:
        """Returns serializable string representation of the class."""
        return f"""{self.__class__.__name__}(
    type = {self.type_},
    recipient = {self.recipient},
    message = {self.message},
    sender = {self.sender},
    sender_type = {self.sender_type},
    econet_version = {self.econet_version},
    data = {self._data}
)
""".strip()

    @property
    def length(self) -> int:
        """Returns frame length.

        Structure:
        HEADER_SIZE bytes
        + 1 bytes frame type
        + length of message
        + 1 byte crc
        + 1 byte end delimiter
        """
        return HEADER_SIZE + 1 + len(self.message) + 1 + 1

    def __len__(self) -> int:
        """Returns frame length when calling len() on class instance."""
        return self.length

    @property
    def data(self):
        """Gets data parsed from frame."""
        if self._data is None:
            # If frame data not present.
            self.parse_message(self.message)

        return self._data

    @property
    def header(self) -> bytearray:
        """Gets frame header as bytearray."""
        buffer = bytearray(HEADER_SIZE)
        util.pack_header(
            buffer,
            0,
            *[
                FRAME_START,
                len(self),
                self.recipient,
                self.sender,
                self.sender_type,
                self.econet_version,
            ],
        )

        return buffer

    def is_type(self, *types: [Frame]) -> bool:
        """Checks if frame belongs to one of specified types.

        Keyword arguments:
        types -- a list of Frame classes to check against
        """
        for type_ in types:
            if isinstance(self, type_):
                return True

        return False

    def to_bytes(self) -> bytes:
        """Converts frame to bytes respresentation."""
        data = self.header
        data.append(self.type_)
        for byte in self.message:
            data.append(byte)

        data.append(util.crc(data))
        data.append(FRAME_END)
        return bytes(data)

    def to_hex(self) -> str:
        """Converts frame to list of hex bytes."""
        return util.to_hex(self.to_bytes())

    @abstractmethod
    def create_message(self) -> bytearray:
        """Creates bytearray message from
        provided data.
        """

    @abstractmethod
    def parse_message(self, message: bytearray) -> None:
        """Parses data from the frame message.

        Keyword arguments:
        message - bytearray message to parse
        """


class Request(Frame):
    """Base class for all requests frames."""

    def response(self, **args) -> Frame:  # pylint: disable=no-self-use
        """Returns instance of Frame
        for response to request, if needed.
        """
        return None

    def create_message(self) -> bytearray:
        """Creates bytearray message from
        provided data.
        """
        return bytearray()

    def parse_message(self, message: bytearray) -> None:
        """Parses data from the frame message.

        Keyword arguments:
        message - bytearray message to parse
        """


class Response(Frame):
    """Base class for all response frames."""

    def create_message(self) -> bytearray:
        """Creates bytearray message from
        provided data.
        """
        return bytearray()

    def parse_message(self, message: bytearray) -> None:
        """Parses data from the frame message.

        Keyword arguments:
        message - bytearray message to parse
        """



"""Contains UID structure parser."""

from pyplumio import util
from pyplumio.constants import UID_BASE, UID_BASE_BITS, UID_CHAR_BITS


def from_bytes(message: bytearray, offset: int = 0) -> (str, int):
    """Parses frame message into usable data.

    Keyword arguments:
    message -- ecoNET message
    offset -- current data offset
    """
    uid_length = message[offset]
    offset += 1
    uid = message[offset : uid_length + offset].decode()
    offset += uid_length
    input_ = uid + util.uid_stamp(uid)
    input_length = len(input_) * UID_CHAR_BITS
    output = []
    output_length = input_length // UID_BASE_BITS
    if input_length % UID_BASE_BITS:
        output_length += 1

    conv_int = 0
    conv_size = 0
    j = 0
    for _ in range(output_length):
        if conv_size < UID_BASE_BITS and j < len(input_):
            conv_int += ord(input_[j]) << conv_size
            conv_size += UID_CHAR_BITS
            j += 1

        char_code = conv_int % UID_BASE
        conv_int //= UID_BASE
        conv_size -= UID_BASE_BITS
        output.insert(0, util.uid_5bits_to_char(char_code))

    return "".join(output), offset



"""Contains outputs structure parser."""

from pyplumio.constants import STATUSES


def from_bytes(message: bytearray, offset: int = 0, data: dict = None) -> (dict, int):
    """Parses frame message into usable data.

    Keyword arguments:
    message -- ecoNET message
    offset -- current data offset
    """
    if data is None:
        data = {}

    for index, status in enumerate(STATUSES):
        data[status] = message[offset + index]

    offset += 4

    return data, offset



"""Contains response frame classes."""

import struct

from pyplumio import util
from pyplumio.constants import (
    DATA_FAN_POWER,
    DATA_FUEL_CONSUMPTION,
    DATA_FUEL_LEVEL,
    DATA_LOAD,
    DATA_MODE,
    DATA_POWER,
    DATA_THERMOSTAT,
    DATA_TRANSMISSION,
    DEFAULT_IP,
    DEFAULT_NETMASK,
    WLAN_ENCRYPTION,
    WLAN_ENCRYPTION_NONE,
)
from pyplumio.frame import Response
from pyplumio.structures import (
    alarms,
    frame_versions,
    lambda_,
    mixer_parameter,
    mixers,
    modules,
    output_flags,
    outputs,
    regulator_parameter,
    statuses,
    temperatures,
    thermostats,
    uid,
    var_string,
)
from pyplumio.version import __version__


class ProgramVersion(Response):
    """Contains information about device software and hardware version."""

    type_: int = 0xC0

    _defaults: dict = {
        "version": __version__,
        "struct_tag": b"\xFF\xFF",
        "struct_version": 5,
        "device_id": b"\x7A\x00",
        "processor_signature": b"\x00\x00\x00",
    }

    def create_message(self) -> bytearray:
        """Creates ProgramVersion message."""
        data = util.merge(self._defaults, self._data)
        version = data["version"].split(".")
        message = bytearray(15)
        struct.pack_into(
            "<2sB2s3sHHHB",
            message,
            0,
            data["struct_tag"],
            data["struct_version"],
            data["device_id"],
            data["processor_signature"],
            *map(int, version),
            self.sender,
        )

        return message

    def parse_message(self, message: bytearray):
        """Parses ProgramVersion message into usable data.

        Keywords arguments:
        message -- message to parse
        """
        self._data = {}
        [
            self._data["struct_tag"],
            self._data["struct_version"],
            self._data["device_id"],
            self._data["processor_signature"],
            version1,
            version2,
            version3,
            self._data["address"],
        ] = struct.unpack_from("<2sB2s3sHHHB", message)
        self._data["version"] = ".".join(map(str, [version1, version2, version3]))


class DeviceAvailable(Response):
    """Contains ecoNET device information."""

    type_: int = 0xB0

    _defaults: dict = {
        "eth": {
            "ip": DEFAULT_IP,
            "netmask": DEFAULT_NETMASK,
            "gateway": DEFAULT_IP,
            "status": False,
        },
        "wlan": {
            "ip": DEFAULT_IP,
            "netmask": DEFAULT_NETMASK,
            "gateway": DEFAULT_IP,
            "status": False,
            "encryption": WLAN_ENCRYPTION[WLAN_ENCRYPTION_NONE],
            "quality": 100,
            "ssid": "",
        },
        "server": {"status": True},
    }

    def create_message(self) -> bytearray:
        """Creates CheckDevice message."""
        message = bytearray()
        message += b"\x01"
        data = util.merge(self._defaults, self._data)
        eth = data["eth"]
        wlan = data["wlan"]
        server = data["server"]
        for address in ("ip", "netmask", "gateway"):
            message += util.ip_to_bytes(eth[address])

        message.append(eth["status"])
        for address in ("ip", "netmask", "gateway"):
            message += util.ip_to_bytes(wlan[address])

        message.append(server["status"])
        message.append(wlan["encryption"])
        message.append(wlan["quality"])
        message.append(wlan["status"])

        message += b"\x00" * 4
        message.append(len(wlan["ssid"]))
        message += wlan["ssid"].encode("utf-8")

        return message

    def parse_message(self, message: bytearray) -> None:
        """Parses CheckDevice message into usable data.

        Keywords arguments:
        message -- message to parse
        """
        self._data = {"eth": {}, "wlan": {}, "server": {}}
        offset = 1
        for part in ("ip", "netmask", "gateway"):
            self._data["eth"][part] = util.ip_from_bytes(message[offset : offset + 4])
            offset += 4

        self._data["eth"]["status"] = bool(message[offset])
        offset += 1
        for part in ("ip", "netmask", "gateway"):
            self._data["wlan"][part] = util.ip_from_bytes(message[offset : offset + 4])
            offset += 4

        self._data["server"]["status"] = bool(message[offset])
        self._data["wlan"]["encryption"] = int(message[offset + 1])
        self._data["wlan"]["quality"] = int(message[offset + 2])
        self._data["wlan"]["status"] = bool(message[offset + 3])
        offset += 8
        self._data["wlan"]["ssid"] = var_string.from_bytes(message, offset)


class CurrentData(Response):
    """Contains current device state data."""

    type_: int = 0x35

    def parse_message(self, message: bytearray) -> None:
        """Parses CurrentData message into usable data.

        Keywords arguments:
        message -- message to parse
        """
        self._data = {}
        offset = 0
        _, offset = frame_versions.from_bytes(message, offset, self._data)
        self._data[DATA_MODE] = message[offset]
        offset += 1
        _, offset = outputs.from_bytes(message, offset, self._data)
        _, offset = output_flags.from_bytes(message, offset, self._data)
        _, offset = temperatures.from_bytes(message, offset, self._data)
        _, offset = statuses.from_bytes(message, offset, self._data)
        _, offset = alarms.from_bytes(message, offset, self._data)
        self._data[DATA_FUEL_LEVEL] = message[offset]
        self._data[DATA_TRANSMISSION] = message[offset + 1]
        self._data[DATA_FAN_POWER] = util.unpack_float(
            message[offset + 2 : offset + 6]
        )[0]
        self._data[DATA_LOAD] = message[offset + 6]
        self._data[DATA_POWER] = util.unpack_float(message[offset + 7 : offset + 11])[0]
        self._data[DATA_FUEL_CONSUMPTION] = util.unpack_float(
            message[offset + 11 : offset + 15]
        )[0]
        self._data[DATA_THERMOSTAT] = message[offset + 15]
        offset += 16
        _, offset = modules.from_bytes(message, offset, self._data)
        _, offset = lambda_.from_bytes(message, offset, self._data)
        _, offset = thermostats.from_bytes(message, offset, self._data)
        _, offset = mixers.from_bytes(message, offset, self._data)


class UID(Response):
    """Contains device UID."""

    type_: int = 0xB9

    def parse_message(self, message: bytearray) -> None:
        """Parses UID message into usable data.

        Keywords arguments:
        message -- message to parse
        """
        self._data = {}
        offset = 0
        self._data["reg_type"] = message[offset]
        offset += 1
        self._data["reg_prod"] = util.unpack_ushort(message[offset : offset + 2])
        offset += 2
        self._data["UID"], offset = uid.from_bytes(message, offset)
        self._data["reg_logo"] = util.unpack_ushort(message[offset : offset + 2])
        offset += 2
        self._data["reg_img"] = util.unpack_ushort(message[offset : offset + 2])
        offset += 2
        self._data["reg_name"], offset = var_string.from_bytes(message, offset)


class Password(Response):
    """Contains device service password."""

    type_: int = 0xBA

    def parse_message(self, message: bytearray) -> None:
        """Parses ProgramVersion message into usable data.

        Keywords arguments:
        message -- message to parse
        """
        password = message[1:]
        if password:
            self._data = password.decode()


class RegData(Response):
    """Contains current regulator data."""

    type_: int = 0x08
    struct: list = []

    VERSION: str = "1.0"
    DATA_TYPES_LEN: list = [0, 1, 2, 4, 1, 2, 4, 4, 0, 8, 1, -1, -1, 8, 8, 4, 16]
    DATATYPE_UNDEFINED0: int = 0
    DATATYPE_SHORT_INT: int = 1
    DATATYPE_INT: int = 2
    DATATYPE_LONG_INT: int = 3
    DATATYPE_BYTE: int = 4
    DATATYPE_WORD: int = 5
    DATATYPE_DWORD: int = 6
    DATATYPE_SHORT_REAL: int = 7
    DATATYPE_UNDEVINED8: int = 8
    DATATYPE_LONG_REAL: int = 9
    DATATYPE_BOOLEAN: int = 10
    DATATYPE_BCD: int = 11
    DATATYPE_STRING: int = 12
    DATATYPE_INT_64: int = 13
    DATATYPE_UINT_64: int = 14
    DATATYPE_IPv4: int = 15
    DATATYPE_IPv6: int = 16

    def parse_message(self, message: bytearray) -> None:
        """Parses RegData message into usable data.

        Keywords arguments:
        message -- message to parse
        """
        offset = 2
        frame_version = f"{message[offset+1]}.{message[offset]}"
        offset += 2
        self._data = {}
        if frame_version == self.VERSION:
            frame_versions.from_bytes(message, offset, self._data)


class Timezones(Response):
    """Contains device timezone info."""

    type_: int = 0xB6


class Parameters(Response):
    """Contains editable parameters."""

    type_: int = 0xB1

    def parse_message(self, message: bytearray) -> None:
        """Parses Parameters message into usable data.

        Keywords arguments:
        message -- message to parse
        """
        self._data, _ = regulator_parameter.from_bytes(message)


class MixerParameters(Response):
    """Contains current mixers parameters."""

    type_: int = 0xB2

    def parse_message(self, message: bytearray) -> None:
        """Parses Parameters message into usable data.

        Keywords arguments:
        message -- message to parse
        """
        self._data, _ = mixer_parameter.from_bytes(message)


class DataStructure(Response):
    """Contains device data structure."""

    type_: int = 0xD5

    def parse_message(self, message: bytearray) -> None:
        """Parses DataStructure message into usable data.

        Keywords arguments:
        message -- message to parse
        """
        offset = 0
        integrity_blocks_number = util.unpack_ushort(message[offset : offset + 2])
        offset += 2
        self._data = []
        if integrity_blocks_number > 0:
            for _ in range(integrity_blocks_number):
                param_type = message[offset]
                param_id = util.unpack_ushort(message[offset + 1 : offset + 3])
                self._data.append({"id": param_id, "type": param_type})
                offset += 3


class SetParameter(Response):
    """Contains set parameter response."""

    type_: int = 0xB3


class SetMixerParameter(Response):
    """Sets mixer parameter."""

    type_: int = 0xB4


class BoilerControl(Response):
    """Contains boiler control response."""

    type_: int = 0xBB



"""Contains classes for supported devices."""
from __future__ import annotations

from .constants import (
    DATA_FRAMES,
    DATA_MODE,
    DEVICE_DATA,
    DEVICE_PARAMS,
    MODES,
    MODULE_PANEL,
)
from .helpers.base_device import BaseDevice
from .helpers.parameter import Parameter
from .mixers import MixersCollection
from .storage import FrameBucket

ECOMAX_ADDRESS: int = 0x45
ECOSTER_ADDRESS: int = 0x51


class Device(BaseDevice):
    """Device representation."""

    def __init__(self, data: dict = None, parameters: dict = None):
        """Creates device instance.

        Keyword arguments:
        data -- device data
        parameters -- editable parameters
        """
        super().__init__(data, parameters)
        self.product = None
        self.uid = None
        self.password = None
        self.struct = []
        self.bucket = FrameBucket()
        self.mixers = MixersCollection()
        self._is_on = False

    def set_data(self, data: dict) -> None:
        """Sets device data.

        Keyword arguments:
        data -- device immutable attributes
        """
        self.bucket.fill(data[DATA_FRAMES])
        for name, value in data.items():
            if name in DEVICE_DATA:
                self._data[name] = value

    def set_parameters(self, parameters: dict) -> None:
        """Sets device parameters.

        Keyword arguments:
        parameters -- device changeable parameters
        """
        for name, parameter in parameters.items():
            if name in DEVICE_PARAMS:
                self._parameters[name] = Parameter(name, *parameter)

        self._parameters["boiler_control"] = Parameter(
            name="boiler_control", value=int(self.is_on), min_=0, max_=1
        )

    def has_mixers(self) -> bool:
        """Check if device instance has mixers."""
        return bool(self.mixers.mixers)

    @property
    def software(self) -> str | None:
        """Returns software version."""
        if MODULE_PANEL in self._data:
            return self._data[MODULE_PANEL]

        return None

    @property
    def is_on(self) -> bool:
        """Returns current state."""
        if DATA_MODE in self._data:
            return bool(self._data[DATA_MODE] != 0)

        return False

    @property
    def mode(self) -> str:
        """Returns current mode."""
        if self._data:
            mode = self._data[DATA_MODE]
            try:
                return MODES[mode]
            except IndexError:
                pass

        return "Unknown"

    @property
    def changes(self) -> list[Parameter]:
        """Returns changed device parameters."""
        changes = self.queue
        changes.extend(self.bucket.queue)
        changes.extend(self.mixers.queue)
        return changes

    @property
    def editable_parameters(self) -> list:
        """Returns list of editable parameters."""
        return DEVICE_PARAMS

    def __str__(self) -> str:
        """Converts device instance to a string."""
        return f"""
Product:        {self.product}
Software Ver.:  {self.software}
UID:            {self.uid}
Password:       {self.password}

{super().__str__()}

Mixers:
{self.mixers}
""".strip()


class EcoMAX(Device):
    """ecoMAX device representation."""

    address = ECOMAX_ADDRESS


class EcoSTER(Device):
    """ecoSTER device representation."""

    address = ECOSTER_ADDRESS


class DevicesCollection:
    """Collection of ecoNET devices."""

    def __init__(self):
        """Creates device collection."""
        self._classes = Device.__subclasses__()
        self._addresses = [v.address for v in self._classes]
        self._instances = {}
        self._names = {}

    def __getattr__(self, name: str) -> Device | None:
        """Gets device by name.""

        Keyword arguments:
        name -- name of device to get
        """
        name = name.lower()
        if name in self._names:
            return self._instances[self._names[name]]

        return None

    def __len__(self) -> int:
        """Returns number of devices in collection."""
        return len(self._instances)

    def has(self, needle) -> bool:
        """Checks if collection has device for specified address.

        Keyword arguments:
        needle -- address or name of device
        """
        return (needle in self._addresses) or (needle in self._names)

    def get(self, address: int) -> Device | None:
        """Gets device by address.

        Keyword arguments:
        address -- int address of device
        """
        if address not in self._instances:
            try:
                index = self._addresses.index(address)
                cls = self._classes[index]
            except ValueError:
                return None

            self._instances[address] = cls()
            self._names[cls.__name__.lower()] = address

        return self._instances[address]



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

pkg_vars = {}
exec(open("pyplumio/version.py").read(), pkg_vars)

setuptools.setup(
    name="PyPlumIO",
    version=pkg_vars["__version__"],
    author="Denis Paavilainen",
    author_email="denpa@denpa.pro",
    description="PyPlumIO is a native ecoNET library for Plum ecoMAX controllers.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/denpamusic/PyPlumIO",
    project_urls={
        "Bug Tracker": "https://github.com/denpamusic/PyPlumIO/issues",
    },
    install_requires=[
        "pyserial-asyncio",
    ],
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Software Development :: Libraries",
    ],
    packages=setuptools.find_packages(exclude=["tests", "tests.*"]),
    python_requires=">=3.8",
)



"""Contains mixers structure parser."""

from pyplumio import util
from pyplumio.constants import DATA_MIXERS, MIXER_PUMP, MIXER_TARGET, MIXER_TEMP


def from_bytes(message: bytearray, offset: int = 0, data: dict = None) -> (list, int):
    """Parses frame message into usable data.

    Keyword arguments:
    message -- ecoNET message
    offset -- current data offset
    """
    if data is None:
        data = {}

    mixers = []
    mixers_number = message[offset]
    offset += 1
    if mixers_number > 0:
        for _ in range(1, mixers_number + 1):
            mixer = {}
            mixer[MIXER_TEMP] = util.unpack_float(message[offset : offset + 4])[0]
            mixer[MIXER_TARGET] = message[offset + 4]
            mixer_outputs = message[offset + 6]
            mixer[MIXER_PUMP] = bool(mixer_outputs & 0x01)
            mixers.append(mixer)
            offset += 8

    data[DATA_MIXERS] = mixers

    return data, offset



"""Contains versions structure parser."""

import struct

from pyplumio.constants import MODULE_PANEL, MODULES


def from_bytes(message: bytearray, offset: int = 0, data: dict = None) -> (dict, int):
    """Parses frame message into usable data.

    Keyword arguments:
    message -- ecoNET message
    offset -- current data offset
    """
    if data is None:
        data = {}

    for module in MODULES:
        if module == MODULE_PANEL:
            version_data = struct.unpack("<BBBBB", message[offset : offset + 5])
            version1 = ".".join(map(str, version_data[:3]))
            version2 = "." + chr(version_data[3])
            version3 = str(version_data[4])
            data[module] = version1 + version2 + version3
            offset += 5
            continue

        if message[offset] == 0xFF:
            data[module] = None
            offset += 1
        else:
            data[module] = ".".join(
                map(str, struct.unpack("<BBB", message[offset : offset + 3]))
            )
            offset += 3

    return data, offset



"""Contains main ecoNET class, that is used to interact with ecoNET
connection.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
import asyncio
from collections.abc import Callable
import logging
import os
import sys

import serial_asyncio

from .constants import (
    DATA_MIXERS,
    DEFAULT_IP,
    DEFAULT_NETMASK,
    READER_TIMEOUT,
    RECONNECT_TIMEOUT,
    WLAN_ENCRYPTION,
)
from .devices import DevicesCollection
from .exceptions import ChecksumError, FrameTypeError, LengthError
from .frame import Frame
from .frames import requests, responses
from .stream import FrameReader, FrameWriter

_LOGGER = logging.getLogger(__name__)


class EcoNET(ABC):
    """Base ecoNET connection class."""

    def __init__(self, **kwargs):
        """Creates EcoNET connection instance.

        Keyword arguments:
        **kwargs -- keyword arguments directly passed to asyncio's
            create_connection method
        """
        self.kwargs = kwargs
        self.closing = False
        self._net = {}
        self._devices = DevicesCollection()
        self._callback_task = None
        self.writer = None

    def __enter__(self):
        """Provides entry point for context manager."""
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Provides exit point for context manager."""

    async def _callback(
        self, callback: Callable[DevicesCollection, EcoNET], interval: int
    ) -> None:
        """Calls provided callback method with specified interval.

        Keyword arguments:
        callback -- callback method
        ecomax -- ecoMAX device instance
        interval -- update interval in seconds
        """
        while True:
            await callback(self._devices, econet=self)
            await asyncio.sleep(interval)

    async def _process(self, frame: Frame, writer: FrameWriter) -> None:
        """Processes received frame.

        Keyword arguments:
        frame -- instance of received frame
        writer -- instance of writer
        """

        if frame is None or not self._devices.has(frame.sender):
            return None

        device = self._devices.get(frame.sender)
        if frame.is_type(requests.ProgramVersion):
            writer.queue(frame.response())

        elif frame.is_type(responses.UID):
            device.uid = frame.data["UID"]
            device.product = frame.data["reg_name"]

        elif frame.is_type(responses.Password):
            device.password = frame.data

        elif frame.is_type(responses.RegData, responses.CurrentData):
            device.set_data(frame.data)
            if DATA_MIXERS in frame.data:
                device.mixers.set_data(frame.data[DATA_MIXERS])

        elif frame.is_type(responses.Parameters):
            device.set_parameters(frame.data)

        elif frame.is_type(responses.MixerParameters):
            device.mixers.set_parameters(frame.data[DATA_MIXERS])

        elif frame.is_type(responses.DataStructure):
            device.struct = frame.data

        writer.collect(device.changes)
        if writer.queue_is_empty():
            # If queue is empty, send device available frame.
            writer.queue(
                responses.DeviceAvailable(recipient=frame.sender, data=self._net)
            )

    async def _read(self, reader: FrameReader) -> None:
        """Handles connection reads."""
        while True:
            if self.closing:
                await self.writer.close()
                self.writer = None
                self.closing = False
                break

            try:
                frame = await asyncio.wait_for(reader.read(), timeout=READER_TIMEOUT)
            except (asyncio.TimeoutError, ConnectionRefusedError, OSError):
                _LOGGER.error(
                    "Connection to device failed, retrying in %i seconds...",
                    RECONNECT_TIMEOUT,
                )
                reader, self.writer = await self.reconnect()
            except ChecksumError:
                _LOGGER.warning("Incorrect frame checksum.")
            except LengthError:
                _LOGGER.warning("Incorrect frame length.")
            except FrameTypeError:
                _LOGGER.info("Unknown frame type %s.", hex(frame.type_))
            else:
                asyncio.create_task(self._process(frame, self.writer))
                await self.writer.process_queue()

    async def reconnect(self) -> (FrameReader, FrameWriter):
        """Initializes reconnect after RECONNECT_TIMEOUT seconds."""
        if self.connected():
            await self.writer.close()

        await asyncio.sleep(RECONNECT_TIMEOUT)
        return await self.connect()

    async def task(
        self, callback: Callable[DevicesCollection, EcoNET], interval: int = 1
    ) -> None:
        """Establishes connection and continuously reads new frames.

        Keyword arguments:
        callback -- user-defined callback method
        interval -- user-defined update interval in seconds
        """
        reader, self.writer = await self.connect()
        self.writer.queue(requests.Password())
        self._callback_task = asyncio.create_task(self._callback(callback, interval))
        await self._read(reader)

    def run(
        self, callback: Callable[DevicesCollection, EcoNET], interval: int = 1
    ) -> None:
        """Run connection in the event loop.

        Keyword arguments:
        callback -- user-defined callback method
        interval -- user-defined update interval in seconds
        """
        if os.name == "nt":
            asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

        try:
            sys.exit(asyncio.run(self.task(callback, interval)))
        except KeyboardInterrupt:
            pass

    def set_eth(
        self, ip: str, netmask: str = DEFAULT_NETMASK, gateway: str = DEFAULT_IP
    ) -> None:
        """Sets eth parameters to pass to devices.
        Used for informational purpoises only.

        Keyword arguments:
        ip -- ip address of eth device
        netmask -- netmask of eth device
        gateway -- gateway address of eth device
        """
        eth = {}
        eth["ip"] = ip
        eth["netmask"] = netmask
        eth["gateway"] = gateway
        eth["status"] = True
        self._net["eth"] = eth

    def set_wlan(
        self,
        ssid: str,
        ip: str,
        encryption: int = WLAN_ENCRYPTION[1],
        netmask: str = DEFAULT_NETMASK,
        gateway: str = DEFAULT_IP,
        quality: int = 100,
    ) -> None:
        """Sets wlan parameters to pass to devices.
        Used for informational purposes only.

        Keyword arguments:
        ssid -- SSID string
        encryption -- wlan encryption, must be passed with constant
        ip -- ip address of wlan device
        netmask -- netmask of wlan device
        gateway -- gateway address of wlan device
        """
        wlan = {}
        wlan["ssid"] = ssid
        wlan["encryption"] = encryption
        wlan["quality"] = quality
        wlan["ip"] = ip
        wlan["netmask"] = netmask
        wlan["gateway"] = gateway
        wlan["status"] = True
        self._net["wlan"] = wlan

    def connected(self) -> bool:
        """Returns connection state."""
        return self.writer is not None

    def close(self) -> None:
        """Closes opened connection."""
        if not self.closing:
            self.closing = True
            if self._callback_task is not None:
                self._callback_task.cancel()

    @abstractmethod
    async def connect(self) -> (FrameReader, FrameWriter):
        """Initializes connection and returns frame reader and writer."""


class TcpConnection(EcoNET):
    """Represents ecoNET TCP connection through."""

    def __init__(self, host: str, port: int, **kwargs):
        """Creates EcoNET TCP connection instance."""
        super().__init__(**kwargs)
        self.host = host
        self.port = port

    def __repr__(self):
        """Creates string respresentation of class."""
        return f"""{self.__class__.__name__}(
    host = {self.host},
    port = {self.port},
    kwargs = {self.kwargs}
)
"""

    async def connect(self) -> (FrameReader, FrameWriter):
        """Initializes connection and returns frame reader and writer."""
        reader, writer = await asyncio.open_connection(
            host=self.host, port=self.port, **self.kwargs
        )
        return [FrameReader(reader), FrameWriter(writer)]


class SerialConnection(EcoNET):
    """Represents ecoNET serial connection."""

    def __init__(self, device: str, baudrate: int = 115200, **kwargs):
        """Creates EcoNET serial connection instance."""
        super().__init__(**kwargs)
        self.device = device
        self.baudrate = baudrate

    def __repr__(self):
        """Creates string respresentation of class."""
        return f"""{self.__class__.__name__}(
    device = {self.device},
    baudrate = {self.baudrate},
    kwargs = {self.kwargs}
)
"""

    async def connect(self) -> (FrameReader, FrameWriter):
        """Initializes connection and returns frame reader and writer."""
        reader, writer = await serial_asyncio.open_serial_connection(
            url=self.device,
            baudrate=self.baudrate,
            bytesize=serial_asyncio.serial.EIGHTBITS,
            parity=serial_asyncio.serial.PARITY_NONE,
            stopbits=serial_asyncio.serial.STOPBITS_ONE,
            **self.kwargs,
        )
        return [FrameReader(reader), FrameWriter(writer)]



"""Contains frame factory class."""

from __future__ import annotations

import inspect

from .exceptions import FrameTypeError
from .frame import Frame
from .frames import requests, responses
from .helpers.singleton import Singleton


class FrameFactory(Singleton):
    """Used to create frame objects based on frame class."""

    def __init__(self) -> None:
        """Calls method to make type list."""
        self._types = {}

    def get_frame(self, type_: int, **kwargs) -> Frame:
        """Gets frame by frame type.

        Keyword arguments:
        type -- integer, repsenting frame type
        kwargs -- keywords arguments to pass to the frame class
        """
        if type_ in self.types:
            return self.types[type_](**kwargs)

        raise FrameTypeError()

    @property
    def types(self) -> dict:
        """Constructs and return a list of available frame
        types and handlers.
        """
        if not self._types:
            for module in [requests, responses]:
                for _, obj in inspect.getmembers(module, inspect.isclass):
                    if obj.__module__ == module.__name__:
                        # Object is within the module.
                        self._types[obj.type_] = obj

        return self._types
